# coding=utf-8

import logging
import sys

import psycopg2.extensions
try:
    from gevent.lock import Semaphore
except ImportError:
    from eventlet.semaphore import Semaphore

try:
    from django.db.backends.postgresql_psycopg2.base import CursorWrapper
except ImportError:
    pass

import django
from django.db.backends.postgresql_psycopg2.base import \
    DatabaseWrapper as OriginalDatabaseWrapper
from django.db.backends.signals import connection_created
from django.conf import settings
from django.db.backends.postgresql_psycopg2.base import utc_tzinfo_factory
from django.utils.encoding import force_str

try:
    import psycopg2_pool as psypool
except ImportError:
    import django_db_geventpool.backends.postgresql_psycopg2.psycopg2_pool as psypool
from .creation import DatabaseCreation

logger = logging.getLogger(__name__)

connection_pools = {}
connection_pools_lock = Semaphore(value=1)


class DatabaseWrapperMixin15(object):
    def __init__(self, *args, **kwargs):
        self._pool = None
        super(DatabaseWrapperMixin15, self).__init__(*args, **kwargs)
        self.creation = DatabaseCreation(self)

    @property
    def pool(self):
        if self._pool is not None:
            return self._pool
        connection_pools_lock.acquire()
        if not self.alias in connection_pools:
            self._pool = psypool.PostgresConnectionPool(
                **self.get_connection_params())
            connection_pools[self.alias] = self._pool
        else:
            self._pool = connection_pools[self.alias]
        connection_pools_lock.release()
        return self._pool

    def get_connection_params(self):
        settings_dict = self.settings_dict
        if not settings_dict['NAME']:
            from django.core.exceptions import ImproperlyConfigured
            raise ImproperlyConfigured(
                "settings.DATABASES is improperly configured. "
                "Please supply the NAME value.")
        conn_params = {
            'database': settings_dict['NAME'],
        }
        conn_params.update(settings_dict['OPTIONS'])
        if 'autocommit' in conn_params:
            del conn_params['autocommit']
        if 'isolation_level' in conn_params:
            del conn_params['isolation_level']
        if settings_dict['USER']:
            conn_params['user'] = settings_dict['USER']
        if settings_dict['PASSWORD']:
            conn_params['password'] = force_str(settings_dict['PASSWORD'])
        if settings_dict['HOST']:
            conn_params['host'] = settings_dict['HOST']
        if settings_dict['PORT']:
            conn_params['port'] = settings_dict['PORT']
        if 'MAX_CONNS' in settings_dict['OPTIONS']:
            conn_params['MAX_CONNS'] = settings_dict['OPTIONS']['MAX_CONNS']
        return conn_params

    def _cursor(self):
        if self.connection is None:
            self.connection = self.pool.get()
            tz = 'UTC' if settings.USE_TZ else self.settings_dict.get('TIME_ZONE')
            if tz:
                try:
                    get_parameter_status = self.connection.get_parameter_status
                except AttributeError:
                    # psycopg2 > 2.0.12 doesn't support get_parameter_status
                    conn_tz = None
                else:
                    conn_tz = get_parameter_status('TimeZone')
                if conn_tz != tz:
                    self.connection.set_isolation_level(
                        psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT)
                    self.connection.cursor().execute(
                        self.ops.set_time_zone_sql(), [tz])
            self.connection.set_isolation_level(self.isolation_level)
            self._get_pg_version()
            connection_created.send(sender=self.__class__, connection=self)
        cursor = self.connection.cursor()
        cursor.tzinfo_factory = utc_tzinfo_factory if settings.USE_TZ else None
        return CursorWrapper(cursor)

    def close(self):
        self.validate_thread_sharing()
        if self.connection is None:
            return  # no need to close anything
        try:
            if self.connection.closed:
                logger.warning(
                    'psycopg2 connections will be reset.')
                self.pool.closeall()
                self.connection = None
            else:
                self.pool.put(self.connection)
                self.connection = None
        except:
            # In some cases (database restart, network connection lost etc...)
            # the connection to the database is lost without giving Django a
            # notification. If we don't set self.connection to None, the error
            # will occur a every request.
            self.connection = None
            logger.warning(
                'psycopg2 error while closing the connection.',
                exc_info=sys.exc_info())
            raise

    def closeall(self):
        for pool in connection_pools.values():
            pool.closeall()


class DatabaseWrapperMixin16(object):
    def __init__(self, *args, **kwargs):
        self._pool = None
        super(DatabaseWrapperMixin16, self).__init__(*args, **kwargs)
        self.creation = DatabaseCreation(self)

    @property
    def pool(self):
        if self._pool is not None:
            return self._pool
        connection_pools_lock.acquire()
        if not self.alias in connection_pools:
            self._pool = psypool.PostgresConnectionPool(
                **self.get_connection_params())
            connection_pools[self.alias] = self._pool
        else:
            self._pool = connection_pools[self.alias]
        connection_pools_lock.release()
        return self._pool

    def get_new_connection(self, conn_params):
        if self.connection is None:
            self.connection = self.pool.get()
        return self.connection

    def get_connection_params(self):
        conn_params = super(DatabaseWrapperMixin16, self).get_connection_params()
        if 'MAX_CONNS' in self.settings_dict['OPTIONS']:
            conn_params['MAX_CONNS'] = self.settings_dict['OPTIONS']['MAX_CONNS']
        return conn_params

    def close(self):
        self.validate_thread_sharing()
        if self.closed_in_transaction or self.connection is None:
            return  # no need to close anything
        try:
            self._close()
        except:
            # In some cases (database restart, network connection lost etc...)
            # the connection to the database is lost without giving Django a
            # notification. If we don't set self.connection to None, the error
            # will occur at every request.
            self.connection = None
            logger.warning(
                'psycopg2 error while closing the connection.',
                exc_info=sys.exc_info())
            raise
        finally:
            self.set_clean()

    def _close(self):
        if self.connection.closed:
            self.pool.closeall()
        else:
            with self.wrap_database_errors:
                self.pool.put(self.connection)
        self.connection = None

    def closeall(self):
        for pool in connection_pools.values():
            pool.closeall()


class DatabaseWrapperMixin17(object):
    def set_clean(self):
        if self.in_atomic_block:
            self.closed_in_transaction = True
            self.needs_rollback = True


if django.VERSION >= (1, 7):
    class DatabaseWrapperMixin(DatabaseWrapperMixin17, DatabaseWrapperMixin16):
        pass
elif django.VERSION >= (1, 6):
    class DatabaseWrapperMixin(DatabaseWrapperMixin16):
        pass
elif django.VERSION >= (1, 4):
    class DatabaseWrapperMixin(DatabaseWrapperMixin15):
        pass
else:
    raise ImportError("Django version 1.4.x or greater needed")


class DatabaseWrapper(DatabaseWrapperMixin, OriginalDatabaseWrapper):
    pass



# coding=utf-8

from functools import wraps

from django.core.signals import request_finished


def close_connection(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        result = f(*args, **kwargs)
        request_finished.send(sender='greenlet')
        return result
    return wrapper



# coding=utf-8

import gevent

from django.utils.unittest import TestCase

from .models import TestModel

from django_db_geventpool.utils import close_connection


@close_connection
def test_multiple_connections(count):
    print 'Test {0} starts'.format(count)
    for x in range(0, 20):
        assert len(TestModel.objects.all()) == 1
    print 'Test {0} ends'.format(count)


class ModelTest(TestCase):
    def test_model_save(self):

        data = 'testing save'
        obj = TestModel.objects.create(data=data)

        obj2 = TestModel.objects.get(pk=obj.pk)
        self.assertEqual(obj.data, obj2.data)

    def test_connections(self):
        TestModel.objects.create(data='test')
        greenlets = []

        for x in range(0, 50):
            greenlets.append(gevent.spawn(test_multiple_connections, x))
        gevent.joinall(greenlets)



# coding=utf-8

# this file is a modified version of the psycopg2 used at gevent examples
# to be compatible with django, also checks if
# DB connection is closed and reopen it:
# https://github.com/surfly/gevent/blob/master/examples/psycopg2_pool.py
import logging
logger = logging.getLogger('django')

try:
    from gevent import queue
except ImportError:
    from eventlet import queue

from psycopg2 import connect, DatabaseError


class DatabaseConnectionPool(object):
    def __init__(self, maxsize=100):
        try:
            if not isinstance(maxsize, (int, long)):
                raise TypeError('Expected integer, got %r' % (maxsize, ))
        except Exception as ex:
            if not isinstance(maxsize, (int,)):
                raise TypeError('Expected integer, got %r' % (maxsize, ))
        self.maxsize = maxsize
        self.pool = queue.Queue(maxsize=maxsize)
        self.size = 0

    def get(self):
        pool = self.pool
        if self.size >= self.maxsize or pool.qsize():
            new_item = pool.get()
            try:
                # check connection is still valid
                self.check_usable(new_item)
                logger.debug("DB connection reused")
            except DatabaseError:
                logger.debug("DB connection was closed, creating new one")
                new_item = self.create_connection()
            return new_item
        else:
            self.size += 1
            try:
                new_item = self.create_connection()
                logger.debug("DB connection created")
            except:
                self.size -= 1
                raise
            return new_item

    def put(self, item):
        try:
           self.pool.put(item, timeout=2)
        except queue.Full:
            item.close()

    def closeall(self):
        while not self.pool.empty():
            conn = self.pool.get_nowait()
            try:
                conn.close()
            except Exception:
                pass
        self.size = 0


class PostgresConnectionPool(DatabaseConnectionPool):
    def __init__(self, *args, **kwargs):
        self.connect = kwargs.pop('connect', connect)
        self.connection = None
        maxsize = kwargs.pop('MAX_CONNS', 4)
        self.args = args
        self.kwargs = kwargs
        DatabaseConnectionPool.__init__(
            self, maxsize)

    def create_connection(self):
        conn = self.connect(*self.args, **self.kwargs)
        # set correct encoding
        conn.set_client_encoding('UTF8')
        return conn

    def check_usable(self, connection):
        connection.cursor().execute('SELECT 1')




# coding=utf-8

from django_db_geventpool.backends.postgresql_psycopg2.base import DatabaseWrapperMixin

from django.contrib.gis.db.backends.postgis.base import \
    DatabaseWrapper as OriginalDatabaseWrapper

class DatabaseWrapper(DatabaseWrapperMixin, OriginalDatabaseWrapper):
    pass



# coding=utf-8

import django

from django.db.backends.postgresql_psycopg2.creation import DatabaseCreation as OriginalDatabaseCreation


class DatabaseCreationMixin16(object):
    def _create_test_db(self, verbosity, autoclobber):
        self.connection.closeall()
        return super(DatabaseCreationMixin16, self)._create_test_db(verbosity, autoclobber)

    def _destroy_test_db(self, test_database_name, verbosity):
        self.connection.closeall()
        return super(DatabaseCreationMixin16, self)._destroy_test_db(test_database_name, verbosity)


class DatabaseCreationMixin17(object):
    def _create_test_db(self, verbosity, autoclobber):
        self.connection.closeall()
        return super(DatabaseCreationMixin17, self)._create_test_db(verbosity, autoclobber)

    def _destroy_test_db(self, test_database_name, verbosity):
        self.connection.closeall()
        return super(DatabaseCreationMixin17, self)._destroy_test_db(test_database_name, verbosity)


if django.VERSION >= (1, 7):
    class DatabaseCreationMixin(DatabaseCreationMixin17):
        pass
else:
    class DatabaseCreationMixin(DatabaseCreationMixin16):
        pass


class DatabaseCreation(DatabaseCreationMixin, OriginalDatabaseCreation):
    pass



# coding=utf-8

from setuptools import setup, find_packages

setup(
    name='django-db-geventpool',
    version='1.20',
    install_requires=[
        'django>=1.5',
        'psycopg2>=2.5.1',
        'psycogreen>=1.0'],
    url='https://github.com/jneight/django-db-geventpool',
    description='Add a DB connection pool using gevent to django',
    long_description=open("README.rst").read(),
    packages=find_packages(),
    include_package_data=True,
    license='Apache 2.0',
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
    ],
    author='Javier Cordero Martinez',
    author_email='jcorderomartinez@gmail.com'
)



# coding=utf-8

from django.db import models
from django.utils.encoding import python_2_unicode_compatible


@python_2_unicode_compatible
class TestModel(models.Model):
    data = models.CharField(max_length=32, blank=True)

    def __str__(self):
        return str(self.pk)
